// Take the file with sign functions and simplifies it.

import java.io.*;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.sax.*;
import javax.xml.transform.stream.*;
import org.w3c.dom.*;
import org.xml.sax.*;

class FunctionSimplify {

	private static void simplify(String in, String out) {
		// get parser
		DocumentBuilder builder;
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setValidating(false);
			factory.setNamespaceAware(true);
			builder = factory.newDocumentBuilder();
			builder.setErrorHandler(new XmlErrorHandler());
		} catch (ParserConfigurationException e) {
			System.err.println(e.getMessage());
			return;
		}
		// parse source
		Document source = null;
		try {
			source = builder.parse(in);
		} catch (IOException e) {
			System.err.println(e.getMessage());
			return;
		} catch (SAXException e) {
			System.err.println(e.getMessage());
			return;
		}
		TreeMap<String,TreeSet<String>> replaces = new TreeMap<String,TreeSet<String>>();
		TreeMap<String,TreeMap<String,TreeSet<Record>>> table = analyze(source, replaces);
		resolveUseas(table);
		resolveUseasSubstring(table, replaces);
		// building target
		Document target = null;
		try {
			target = builder.newDocument();
			construct(table, target);
			TransformerFactory factory = SAXTransformerFactory.newInstance();
			factory.setAttribute("indent-number", new Integer(4));
			Transformer transformer = factory.newTransformer();
			transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			DOMSource targetDom = new DOMSource(target);
			File file = new File(out);
			Result targetResult = new StreamResult(file);
			transformer.transform(targetDom, targetResult);
		} catch (Exception e) {
			System.err.println(e);
		}
	}

	// Analyze source.
	private static TreeMap<String,TreeMap<String,TreeSet<Record>>> 
						analyze(Document source,
								TreeMap<String,TreeSet<String>> replaces) {
		TreeMap<String,TreeMap<String,TreeSet<Record>>> table =
			new TreeMap<String,TreeMap<String,TreeSet<Record>>>();
		for (Node child = source.getFirstChild().getFirstChild();
						child != null; 
						child = child.getNextSibling()) {
			if (child instanceof Element) {
				Element el = (Element) child;
				String className = el.getNodeName();
				if (table.get(className) == null)
					table.put(className, new TreeMap<String,TreeSet<Record>>());
				TreeMap<String,TreeSet<Record>> classTable =
					table.get(className);
				Element hi = (Element) el.getElementsByTagName("hi").item(0);
				Text hiText = (Text) hi.getChildNodes().item(0);
				String hiString = linearize(hiText.getData());
				if (classTable.get(hiString) == null)
					classTable.put(hiString, new TreeSet<Record>());
				TreeSet<Record> values = classTable.get(hiString);

				if (className.equals("det-descr")) { 
					NodeList descrList = el.getElementsByTagName("descr");
					if (descrList.getLength() > 0) {
						Element descr = (Element) descrList.item(0);
						Text descrText = (Text) descr.getChildNodes().item(0);
						String descrString = descrText.getData();
						values.add(new DescrRecord(descrString));
					} else
						System.err.println("det-descr without descr");
				} else if (className.equals("typ")) { 
					NodeList descrList = el.getElementsByTagName("descr");
					if (descrList.getLength() > 0) {
						Element descr = (Element) descrList.item(0);
						Text descrText = (Text) descr.getChildNodes().item(0);
						String descrString = descrText.getData();
						values.add(new DescrRecord(descrString));
					} else
						System.err.println("typ without descr");
				} else if (!className.equals("useas")) {
					NodeList alList = el.getElementsByTagName("al");
					if (alList.getLength() > 0) {
						Element al = (Element) alList.item(0);
						Text alText = (Text) al.getChildNodes().item(0);
						String alString = alText.getData();
						String root = al.getAttribute("root");
						addInflected(values, alString, root);
					} else 
						values.add(new DescrRecord(""));
				} else {
					String alt = el.getAttribute("alt");
					String num = el.getAttribute("num");
					try {
						String repeated = "";
						int n = Integer.parseInt(num);
						for (int i = 0; i < n; i++) {
							if (i == 0) 
								repeated += alt;
							else 
								repeated += "-" + alt;
						}
						if (n > 1)
							values.add(new HieroRecord(repeated));
						else {
							if (replaces.get(alt) == null)
								replaces.put(alt, new TreeSet<String>());
							replaces.get(alt).add(hiString);
						}
					} catch (NumberFormatException e) {
					}
				}
			}
		}
		return table;
	}

	// Add transliterations and inflected forms.
	private static void addInflected(TreeSet<Record> translits, String full, String root) {
		translits.add(new TransRecord(full, full, false));
		if (root != null && !root.equals("")) {
			if (full.endsWith("yt") && root.length() <= full.length()-2) {
				// e.g. for full=jHyt, root=jH, add jHwt.
				String inflected = full.substring(0, full.length()-2) + "wt";
				translits.add(new TransRecord(inflected, full, true));
			} else if (full.endsWith("t") && root.length() <= full.length()-1) {
				// e.g. for full=spt, root=sp, add spwt.
				String inflected = full.substring(0, full.length()-1) + "wt";
				translits.add(new TransRecord(inflected, full, true));
			}
			for (int i = root.length(); i < full.length(); i++) {
				// e.g. for full=jAwj, root=jA, add jA and jAw
				String inflected = full.substring(0, i);
				if (!inflected.equals(full))
					translits.add(new TransRecord(inflected, full, false));
			}
		}
	}

	// Pattern of sign name.
	private static Pattern signPattern =
		Pattern.compile("([A-IK-Z]|NL|NU|Aa)[1-9][0-9]*[a-z]?");

	// Turn into simple sequence of signs.
	private static String linearize(String hi) {
		String linear = "";
		Matcher signMatcher = signPattern.matcher(hi);
		while (signMatcher.find()) 
			linear +=
				(linear.equals("") ? "" : "-") + 
				signMatcher.group(0);
		return linear;
	}

	// For holding shortcut.
	private static class Shortcut {
		public String hi;
		public String className;
		public Record val;
		public Shortcut(String hi, String className, Record val) {
			this.hi = hi;
			this.className = className;
			this.val = val;
		}
	}

	// Shortcut useas values.
	private static void resolveUseas(TreeMap<String,TreeMap<String,TreeSet<Record>>> table) {
		Vector<Shortcut> shortcuts = new Vector<Shortcut>();
		for (Map.Entry<String,TreeSet<Record>> pair : table.get("useas").entrySet()) {
			String hi = pair.getKey();
			for (Record record : pair.getValue()) {
				HieroRecord alt = (HieroRecord) record;
				for (Map.Entry<String,TreeMap<String,TreeSet<Record>>> 
						classPair : table.entrySet()) {
					String className = classPair.getKey();
					TreeMap<String,TreeSet<Record>> classTable = classPair.getValue();
					if (!className.equals("useas")) {
						TreeSet<Record> vals = classTable.get(alt.hi);
						if (vals != null)
							for (Record val : vals) 
								shortcuts.add(new Shortcut(hi, className, val));
					}
				}
			}
		}
		for (Shortcut shortcut : shortcuts) {
			String hi = shortcut.hi;
			String className = shortcut.className;
			Record val = shortcut.val;
			TreeMap<String,TreeSet<Record>> classTable = table.get(className);
			if (classTable.get(hi) == null)
				classTable.put(hi, new TreeSet<Record>());
			TreeSet<Record> values = classTable.get(hi);
			values.add(val);
		}
	}

	// Shortcut with longer patterns, where substring has useas that is single
	// sign.
	// First make reverse useas table (replaces). 
	// Then substitute in meanings.
	private static void resolveUseasSubstring(
			TreeMap<String,TreeMap<String,TreeSet<Record>>> table,
			TreeMap<String,TreeSet<String>> replaces) {
		Vector<Shortcut> shortcuts = new Vector<Shortcut>();
		for (Map.Entry<String,TreeMap<String,TreeSet<Record>>> 
				classPair : table.entrySet()) {
			String className = classPair.getKey();
			TreeMap<String,TreeSet<Record>> classTable = classPair.getValue();
			if (!className.equals("useas")) 
				for (Map.Entry<String,TreeSet<Record>> 
						signPair : classTable.entrySet()) {
					String signs = signPair.getKey();
					if (signs.split("-").length > 1) {
						TreeSet<Record> values = signPair.getValue();
						for (String variation : variations(signs, replaces)) {
							for (Record val : values)
								shortcuts.add(new Shortcut(variation, className, val));
						}
					}
				}
		}
		for (Shortcut shortcut : shortcuts) {
			String hi = shortcut.hi;
			String className = shortcut.className;
			Record val = shortcut.val;
			TreeMap<String,TreeSet<Record>> classTable = table.get(className);
			if (classTable.get(hi) == null)
				classTable.put(hi, new TreeSet<Record>());
			TreeSet<Record> values = classTable.get(hi);
			values.add(val);
		}
	}

	// For expressions with multiple glyphs, replace in all ways by
	// alternative glyphs.
	private static Vector<String> variations(String hi, 
			TreeMap<String,TreeSet<String>> replaces) {
		Vector<String> vars = new Vector<String>();
		Vector<String> varsNotEqual = new Vector<String>();
		String[] glyphs = hi.split("-");
		variations(glyphs, 0, replaces, vars);
		for (String var : vars) 
			if (!var.equals(hi)) {
				varsNotEqual.add(var);
			}
		return varsNotEqual;
	}
	private static void variations(String[] glyphs, int i, 
			TreeMap<String,TreeSet<String>> replaces,
			Vector<String> vars) {
		if (i >= glyphs.length) {
			vars.add(join(glyphs));
		} else {
			variations(glyphs, i+1, replaces, vars);
			TreeSet<String> replace = replaces.get(glyphs[i]);
			if (replace != null)
				for (String repl : replace) {
					String[] glyphsVar = new String[glyphs.length];
					System.arraycopy(glyphs, 0, glyphsVar, 0, glyphs.length);
					glyphsVar[i] = repl;
					variations(glyphsVar, i+1, replaces, vars);
				}
		}
	}

	// Join into single string.
	private static String join(String[] glyphs) {
		StringBuffer b = new StringBuffer();
		for (int i = 0; i < glyphs.length; i++) {
			b.append(glyphs[i]);
			if (i < glyphs.length - 1)
				b.append("-");
		}
		return b.toString();
	}

	// Construct target.
	private static void construct(
					TreeMap<String,TreeMap<String,TreeSet<Record>>> table,
				Document target) {
		Element signinfo = target.createElement("signinfo");
		target.appendChild(signinfo);
		for (Map.Entry<String,TreeMap<String,TreeSet<Record>>> 
				classPair : table.entrySet()) {
			String className = classPair.getKey();
			TreeMap<String,TreeSet<Record>> classTable = classPair.getValue();
			for (Map.Entry<String,TreeSet<Record>> 
					signPair : classTable.entrySet()) {
				String sign = signPair.getKey();
				TreeSet<Record> values = signPair.getValue();
				// one empty value cancels all other values
				if (className.equals("det") &&
					   values.contains(new DescrRecord(""))) {
					values = new TreeSet<Record>();
					values.add(new DescrRecord(""));
				}
				for (Record value : values) {
					String elName = className.equals("det-descr") ?
						"det" : className;
					Element function = target.createElement(elName);
					signinfo.appendChild(function);
					Element hi = target.createElement("hi");
					Text hiText = target.createTextNode(sign);
					hi.appendChild(hiText);
					function.appendChild(hi);
					if (className.equals("useas")) {
						Element alt = target.createElement("alt");
						HieroRecord hiValue = (HieroRecord) value;
						Text altText = target.createTextNode(hiValue.hi);
						alt.appendChild(altText);
						function.appendChild(alt);
					} else {
						function.setAttribute("count", "0");
						if (value instanceof DescrRecord) {
							DescrRecord descrRecord = (DescrRecord) value;
							String descr = descrRecord.descr;
							if (!descr.equals("")) {
								Element al = target.createElement("descr");
								Text alText = target.createTextNode(descr);
								al.appendChild(alText);
								function.appendChild(al);
							}
						} else {
							TransRecord transRecord = (TransRecord) value;
							String trans = transRecord.trans;
							String lemma = transRecord.lemma;
							boolean femPlur = transRecord.femPlur;
							Element al = target.createElement("al");
							Text alText = target.createTextNode(trans);
							al.appendChild(alText);
							if (transRecord.femPlur) 
								al.setAttribute("femplur", "true");
							function.appendChild(al);
							if (!lemma.equals(trans)) {
								Element lemmaEl = target.createElement("lemma");
								Text lemmaText = target.createTextNode(lemma);
								lemmaEl.appendChild(lemmaText);
								function.appendChild(lemmaEl);
							}
						}
					}
				}
			}
		}
	}

	// Generic record, for description or transliteration.
	private static abstract class Record implements Comparable<Record> {
		protected abstract String key();
		public int compareTo(Record other) {
			return key().compareTo(other.key());
		}
		public boolean equals(Record other) {
			return other != null &&
				other instanceof Record &&
				key().equals(((Record) other).key());
		}
	}

	// Record with description.
	private static class DescrRecord extends Record {
		public String descr;
		public DescrRecord(String descr) {
			descr = descr.replaceAll("\\s+", " ");
			this.descr = descr;
		}
		protected String key() {
			return "descr " + descr;
		}
	}

	// Record with transliteration, with lemma if applicable, and
	// indication whether feminine plural ending is included.
	private static class TransRecord extends Record {
		public String trans;
		public String lemma;
		public boolean femPlur;
		public TransRecord(String trans, String lemma, boolean femPlur) {
			trans = trans.replaceAll("\\s+", " ");
			lemma = lemma.replaceAll("\\s+", " ");
			this.trans = trans;
			this.lemma = lemma;
			this.femPlur = femPlur;
		}
		public TransRecord(String trans) {
			this(trans, trans, false);
		}
		protected String key() {
			return "trans " + trans + " " + lemma + " " + femPlur;
		}
	}

	// Record with hieroglyphic string.
	private static class HieroRecord extends Record {
		public String hi;
		public HieroRecord(String hi) {
			this.hi = hi;
		}
		protected String key() {
			return "hiero " + hi;
		}
	}

	// The source file is parsed, analysed and
	// simplified and written to another file.
	public static void main(String[] args) {
		if (args.length >= 2) 
			simplify(args[0], args[1]);
	}

}
