#!/usr/bin/perl

# All .html.txt files in the directory are turned into
# .html files, by replacing each hieroglyphic encoding by
# one or more images. There is cashing, to avoid doing this
# several times for the same group of hieroglyphs.
# The mapping from groups to figure numbers is stored in
# reshtml.txt.
# There is a temporary file reshtml.tmp for hieroglyphic encoding.

# Classpath for philolog.
$philo_cp = "philologeg:philologeg/lib/itextpdf-5.4.3.jar";

# Directory where figures are put.
$graphics_dir = "reshtml-graphics";
$index_file = "$graphics_dir/reshtml.txt";
# File where encoding temporarily resides.
$tmp_file = "$graphics_dir/reshtml.tmp";

# Font size for hieroglyphic
$font_size = 30;
# Font size for transliteration.
$translit_font_size = 18;
# Style for transliteration (can be ib, i, sb, s).
$translit_style = "sb";
# Space between lines containing hieroglyphic (in pixels).
$half_leading = 8;

# Create directory for graphics, unless it already exists.
unless (-d $graphics_dir) {
    mkdir($graphics_dir) || die "cannot create $graphics_dir";
}

# If command line argument is "clean", then start with clean slade.
if ($#ARGV == 0 && $ARGV[0] eq "clean") {
    unlink($index_file);
    @trashes = <$graphics_dir/reshtml-*.png>;
    foreach $trash (@trashes) {
	unlink($trash);
    }
}

# Read mapping from encoding to figure numbers.
%encoding_to_figure_num = ();
$max_figure_num = -1;
if (open(MAPPING, $index_file)) {
    @lines = <MAPPING>;
    for ($i = 0; $i < $#lines; $i = $i + 2) {
        $res = $lines[$i];
	$res =~ s/\n//;
        $figure_num = $lines[$i+1];
	$figure_num =~ s/\n//;
	$encoding_to_figure_num{$res} = $figure_num;
	if ($figure_num > $max_figure_num) {
	    $max_figure_num = $figure_num;
	}
    }
    close(MAPPING);
}

# Read files in slurp mode from now on.
undef $/;

# Translate all files with extension html.txt into a html file.
@sources = <*.html.txt>;
foreach $source (@sources) {
    $target = $source;
    $target =~ s/\.html\.txt$/\.html/;
    translate_html($source, $target);
}

# Translate one 'pre' HTML page into one HTML page.
sub translate_html {
    my $source = $_[0];
    my $target = $_[1];
    local($txt);

    open(SOURCE, "$source") || die "cannot open $source";
    $txt = <SOURCE>;
    close(SOURCE);

    open(TARGET, ">$target") || die "cannot open $target";
    $txt = translate_tag($txt, "resmain", 
	"java -cp $philo_cp nederhof/res/ResToImage " .
	"-png -color -fontsize $font_size " .
	"-e \$(pwd)/$tmp_file " .
	"-b \$(pwd)/$graphics_dir/reshtml-RESNUM");
    $txt = translate_tag($txt, "reschunks", 
	"java -cp $philo_cp nederhof/res/ResToImage -chunks " .
	"-png -color -fontsize $font_size " .
	"-e \$(pwd)/$tmp_file " .
	"-b \$(pwd)/$graphics_dir/reshtml-RESNUM-");
    $txt = make_transliteration($txt);
    print TARGET $txt;
    close(TARGET);
}

# Substitute one kind of encoding, for given tag and command.
sub translate_tag {
    my $page = $_[0];
    my $tag = $_[1];
    my $command = $_[2];
    local($open_tag, $close_tag, $out_page, $from_pos, $to_pos);

    $open_tag = "<$tag>";
    $close_tag = "</$tag>";
    $out_page = "";
    $pos = 0;
    while ($pos >= 0) {
	$from_pos = index($page, $open_tag, $pos);
	if ($from_pos >= 0) {
	    $out_page .= substr($page, $pos, $from_pos-$pos);
	    $from_pos += length($open_tag);
	    $to_pos = index($page, $close_tag, $from_pos);
	    if ($to_pos >= 0) {
		$encoding = substr($page, $from_pos,
			$to_pos-$from_pos);
		$encoding =~ s/\n/ /g;
		$out_page .= images_from($encoding, $command, $open_tag);
		$pos = $to_pos + length($close_tag);
	    } else {
		$pos = -1;
	    }
	} else {
	    $out_page .= substr($page, $pos);
	    $pos = -1;
	}
    }
    
    return $out_page;
}

# Produce images for RES encoding.
sub images_from {
    my $res = $_[0];
    my $command = $_[1];
    my $tag = $_[2];
    local($images, $num, $part_num, $figures);

    $images = "";
    if (exists $encoding_to_figure_num{$tag.$res}) {
	$num = $encoding_to_figure_num{$tag.$res};
    } else {
	$max_figure_num++;
	$num = $max_figure_num;
	$encoding_to_figure_num{$tag.$res} = $num;
	$command =~ s/RESNUM/$num/;
	$command .= " > /dev/null";

	open(TMP, ">$tmp_file") || die "cannot open $tmp_file";
	print TMP $encoding;
	close TMP;
	print "Converting to image $res\n";
	system($command);

    }

    if (-e "$graphics_dir/reshtml-$num.png") {
	$images .= image_from("$graphics_dir/reshtml-$num.png");
    } else {
        $part_num = 1;
	while (-e "$graphics_dir/reshtml-$num-$part_num.png") {
	    if ($part_num > 1) {
		$images .= " ";
	    }
	    $images .= image_from("$graphics_dir/reshtml-$num-$part_num.png");
	    $part_num++;
	}
    }
    
    return $images;
}

# Get image for file.
sub image_from {
    my $fig = $_[0];

    return "<img src=\"$fig\" " .
	"style=\"margin-bottom:$half_leading; margin-top:$half_leading;\" " .
	"alt=\"RES\" " .
	"align=\"absmiddle\">";
}

# Translate words of transliteration into images.
sub make_transliteration {
    my $page = $_[0];
    local($open_tag, $close_tag, $out_page, $from_pos, $to_pos);

    $open_tag = "<translit>";
    $close_tag = "</translit>";
    $out_page = "";
    $pos = 0;
    while ($pos >= 0) {
        $from_pos = index($page, $open_tag, $pos);
        if ($from_pos >= 0) {
            $out_page .= substr($page, $pos, $from_pos-$pos);
            $from_pos += length($open_tag);
            $to_pos = index($page, $close_tag, $from_pos);
            if ($to_pos >= 0) {
                $translit = substr($page, $from_pos,
                        $to_pos-$from_pos);
                $translit =~ s/\s+/ /g;
		@words = split(/ /, $translit);
		$words_printed = false;
		for ($j = 0; $j <= $#words; $j++) {
		    $word = $words[$j];
		    if ($word ne "") {
		        if ($words_printed) {
			    $out_page .= " ";
			}
			$out_page .= image_from_word($word);
			$words_printed = true;
		    }
		}
                $pos = $to_pos + length($close_tag);
            } else {
                $pos = -1;
            }
        } else {
            $out_page .= substr($page, $pos);
            $pos = -1;
        }
    }

    return $out_page;
}

# Produce image for word in transliteration.
sub image_from_word {
    my $word = $_[0];
    local(@letters, $name, $c, $char, $filename, $command);
    
    @letters = split(//, $word);
    $name = "";
    for ($c = 0; $c <= $#letters; $c++) {
	$char = $letters[$c];
	if ($char =~ /[a-zA-Z]/) {
	    $name .= $char;
	} else {
	    $name .= "-" . ord($char) . "-";
	}
    }
    $filename = "$graphics_dir/reshtml--$name.png";
    if (!(-e $filename)) {
	$command = "java -cp $philo_cp " .
		"nederhof/images/TransliterationImage " .
		"\"$word\" $filename $translit_font_size $translit_style";
	print "Converting to image $word\n";
	system($command);
    }

    return "<img src=\"$filename\" " .
	"alt=\"transliteration\" " .
	"align=\"absmiddle\">";
}

# Write mapping from encoding to figure numbers.
open(MAPPING, ">$index_file") || die "cannot open $index_file";
foreach $res (keys %encoding_to_figure_num) {
    $figure_num = $encoding_to_figure_num{$res};
    print MAPPING "$res\n";
    print MAPPING "$figure_num\n";
}
close(MAPPING);

# Cleanup tmp file.
unlink("$tmp_file");
